{bot}
{connect4bot is a bot that allows users to play connect4 in channels}
{unlike normal service bots it is not +k it joins when invited and can be kicked}
{like any other user}

(*
 *  beware IRC services, connect4bot.pas
 *  Copyright (C) 2003 Peter Green
 *  Portions Copyright (C) 2002 Bas Steendijk
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit connect4bot;
interface
uses invkickbot,bchannel,buser,bstuff,bparse,bcmds,pgdebugout,pgtypes,sysutils;
const
  //colors are mirc color codes
  boardwidth=7;
  boardheight=6;
  numberinrowtowin = 4;

  colorchar=#$03;
  player1color     = '8';
  player2color     = '4';
  backcolor        = '1';
  tokenchar        = '@';
  emptycolor       = '12';
  emptychar        = '@';
  textcolor        = '0';
  firstcol         = 1;
  firstcolchar     = char(firstcol+ord('0'));
  lastcol          = boardwidth;
  lastcolchar     = char(lastcol+ord('0'));
  lastrow          = boardheight-1;//rows start at 0;
  pl1col=colorchar+player1color;
  pl2col=colorchar+player2color;
  emptcol=colorchar+emptycolor;
  txtcol=colorchar+textcolor;
  backcol=','+backcolor;

type
  tconnect4bot=class(tinvkickbot)
    procedure drawboard(ch:tchannel;won:boolean);
    procedure endgame(ch:tchannel;reason: string);
    // note : userleavechannel only fires for channels the bot is on and
    // does not fire for the bot itself

    procedure userleavechannel(ch:tchannel;user:tuser);override;
    procedure botleavechannel(ch:tchannel);override;
    procedure setupsrvunit;override;
    constructor create;
  end;
  tconnect4game=class
    board : array[firstcol..lastcol,0..lastrow] of byte;
    player1 : tuser;
    player2 : tuser;
    currentplayer : tuser;
    function boardrc(col,row:shortint):byte;
  end;


implementation

uses btime,boptiontable;

var
  {config file options for this service unit}
  locopt:record
    nick:string;
    userid:string;
    host:string;
    info:string;
    maxpenalty:integer;
    penaltypercommand:integer;
  end;

const
  configfile='connect4.ini';
  cfgtag='[connect4]';
  maxoptiontable=5;
  optiontable:array[0..maxoptiontable] of Toptionentry=(
  (
  p:@locopt.nick;
  typ:opt_str;
  name:'nick';
  info:'the nick of this bot';
  def_s:'connect4'),
  (
  p:@locopt.userid;
  typ:opt_str;
  name:'userid';
  info:'the userid of this bot (empty for global userid setting)';
  def_s:'connect4'),
  (
  p:@locopt.host;
  typ:opt_str;
  name:'host';
  info:'the host of this bot (empty for global host setting)';
  def_s:''),
  (
  p:@locopt.info;
  typ:opt_str;
  name:'info';
  info:'this bot''s realname field';
  def_s:'the connect four game bot'),
  (
  p:@locopt.maxpenalty;
  typ:opt_int;
  name:'maxpenalty';
  info:'penalty/throttling setting';
  def_i:60),
  (
  p:@locopt.penaltypercommand;
  typ:opt_int;
  name:'penaltypercommand';
  info:'penalty/throttling setting (seconds per command)';
  def_i:15)
  );

procedure onprivmsg(nickatserver:boolean;sptr:tuser;target:tuser;s:string;tag:taddrint);
begin
  with tconnect4bot(tag) do begin
    s := ircupper(s);
    if s = 'HELP' then begin
      send(mybot.idstr+' '+TOK_NOTICE+' '+sptr.idstr+
      ' :to get '+mybot.name+' on your channel, invite it. to get rid of it, kick it. '+
      'commands, in channel: !start <nick> and !stop to begin/end a game. !1 - !7 to insert coin.');
    end else if s[1] = #1 then begin
      if copy(s,1,6) = #1'PING ' then
      send(mybot.idstr+' '+TOK_NOTICE+' '+sptr.idstr+' :'+s)
    end else begin
      send(mybot.idstr+' '+TOK_NOTICE+' '+sptr.idstr+' :/msg '+mybot.name+' help for help');
    end;
  end;
end;


function spacestring(l: integer): string;
begin
  if L < 1 then begin
    result := '';
  end else begin
    setlength(result,l);
    fillchar(result[1],l,' ');
  end;
end;

procedure tconnect4bot.drawboard(ch:tchannel;won:boolean);
var
  blockwidth : byte; // blockwidth is the width to block out to
                     // so that the background is a nice rectangle
                     // unfortunately the rather heavy useage of mirc colors
                     // makes this slightly harder
  currentline : string;
  row ,col,prevcolor: byte;
  pad : string;
  g : tconnect4game;
begin
  g := tconnect4game(ch.bottag[srvunit.num]);
  if assigned(ch.bottag[srvunit.num]) then with g do begin
    //board line width
    blockwidth := (lastcol-firstcol)*2+3;

    //title line witdth
    if (length(player1.name)+length(player2.name)+6) > blockwidth then blockwidth := length(player1.name)+length(player2.name)+6;

    if assigned(currentplayer) then begin
      if won then begin
        //player wins! line width
        if (length(currentplayer.name)+8) > blockwidth then blockwidth := (length(currentplayer.name)+8);
      end else begin
        //player to move line width
        if (length(currentplayer.name)+10) > blockwidth then blockwidth := (length(currentplayer.name)+10);
      end;
    end else begin
      // board full game drwan line width
      if 23 > blockwidth then blockwidth := 23;
    end;

    currentline := pl1col+backcol+' '+player1.name+txtcol+' VS'+pl2col+' '+player2.name+spacestring(blockwidth-5-length(player1.name)-length(player2.name));
    send(mybot.idstr+' '+tok_privmsg+' '+ch.name+' :'+currentline);

    pad := spacestring(blockwidth-boardwidth*2);
    for row := lastrow downto 0 do begin
      currentline := emptcol+backcol;
      prevcolor := 0;
      for col := firstcol to lastcol do begin
        case board[col,row] of
          0 : begin
            if prevcolor <> 0 then currentline := currentline + emptcol;
            currentline := currentline +' '+emptychar;
          end;
          1 : begin
            if prevcolor <> 1 then currentline := currentline + pl1col;
            currentline := currentline +' '+tokenchar;
          end;
          2 : begin
            if prevcolor <> 2 then currentline := currentline + pl2col;
            currentline := currentline +' '+tokenchar;
          end;
        end;
        prevcolor := board[col,row];
      end;
      currentline := currentline + pad;
      send(mybot.idstr+' '+tok_privmsg+' '+ch.name+' :'+currentline);
    end;
    currentline := txtcol+backcol;
    for col := firstcol to lastcol do begin
      currentline := currentline+' '+char(col+ord('0'));
    end;
    currentline := currentline+pad ;
    send(mybot.idstr+' '+tok_privmsg+' '+ch.name+' :'+currentline);
    if assigned(currentplayer) then begin
      if currentplayer=player1 then begin
        currentline := pl1col+backcol;
      end else begin
        currentline := pl2col+backcol;
      end;
      if won then begin
        currentline := currentline+' '+currentplayer.name+' wins! '+spacestring(blockwidth-length(currentplayer.name)-8);
      end else begin
        currentline := currentline+' '+currentplayer.name+' to move '+spacestring(blockwidth-length(currentplayer.name)-10);
      end;
    end else begin
      currentline := txtcol+backcol+' board full game drawn '+spacestring(blockwidth-23);
    end;
    send(mybot.idstr+' '+tok_privmsg+' '+ch.name+' :'+currentline);
  end;
end;

function tconnect4game.boardrc(col,row:shortint):byte;  //returns value from
                                                        //board or 255 if params
                                                        // out of range
begin
  if (col < firstcol)or(col > lastcol)or(row < 0)or(row > lastrow) then begin
    result := 255;
  end else begin
    result := board[col,row];
  end;
end;

procedure tconnect4bot.endgame(ch:tchannel;reason: string);
begin

  if assigned(ch.bottag[srvunit.num]) then begin
    tconnect4game(ch.bottag[srvunit.num]).free;
    ch.bottag[srvunit.num] := nil;
    send(mybot.idstr+' '+tok_privmsg+' '+ch.name+' :game ended because '+reason);
  end;
end;


procedure onchanmsg(sptr:tuser;target:tchannel;s:string;tag:taddrint);
const
  changerowcoltable:array[0..7,0..1] of integer=(
  (1,0),(1,1),(0,1),(-1,1),(-1,0),(-1,-1),(0,-1),(1,-1));

var
  m2                : string;
  index             : integer;
  cmd               : string;
  localp1 , localp2 : tuser;
  p2name            : string;
  won               : boolean;
  row     , col     : shortint;
  zerofound         : boolean;
  currentplayernum  : byte;
  counter           : byte;
  //innercounter      : byte;
  changerow         : shortint;
  changecol         : shortint;
  wingrid           : array[-1..1,-1..1]of byte;
  g                 : tconnect4game;
function checkpenalty:boolean;
var
  a:integer;
  ts:^integer;
begin
  with tconnect4bot(tag) do begin
    ts := @sptr.bottag[srvunit.num];

    result := false;
    if ts^ < unixtime then ts^ := unixtime;
    a := unixtime+locopt.maxpenalty;
    if ts^ > a then begin
      send(mybot.idstr+' '+TOK_NOTICE+' '+sptr.idstr+' :please wait '+inttostr(integer(sptr.bottag[srvunit.num])-a)+' seconds.');
      result := true;
      inc(ts^,2);
      exit;
    end;
    inc(ts^,locopt.penaltypercommand);
  end;
end;

begin
  with tconnect4bot(tag) do begin
    if length(s) < 2 then exit;
    if s[1] <> '!' then exit;
    m2 := ircupper(s);
    case m2[2] of
      'S' : begin
        index := 2;
        strtok2(m2,' ',index,cmd);
        debugout(cmd);
        if cmd='START' then begin
          if not assigned(target.bottag[srvunit.num]) then begin
            if checkpenalty then exit;
            localp1 := sptr;
            strtok2(m2,' ',index,p2name);
            localp2 := findname(p2name);
            if not assigned(getuserchan(localp1,target)) then begin
              exit;
            end;

            if not assigned(getuserchan(localp2,target)) then begin
              send(mybot.idstr+' '+tok_privmsg+' '+target.name+' :other player not found');
              exit;
            end;
            if localp1=localp2 then begin
              send(mybot.idstr+' '+tok_privmsg+' '+target.name+' :you can''t play yourself');
              exit;
            end;
            target.bottag[srvunit.num] := tconnect4game.Create;
            with tconnect4game(target.bottag[srvunit.num]) do begin
              player1 := localp1;
              player2 := localp2;
              currentplayer := player1;
              drawboard(target,false);
            end;
          end else begin
            send(mybot.idstr+' '+tok_privmsg+' '+target.name+' :a game is already in progress')
          end;
        end else if cmd='STOP' then begin
          if assigned(target.bottag[srvunit.num]) then with tconnect4game(target.bottag[srvunit.num]) do begin
            if (sptr=player1) or (sptr=player2) or isoper(sptr) or hasops(sptr,target,nil) then begin
              endgame(target,sptr.name+' used !STOP')
            end else begin
              send(mybot.idstr+' '+tok_privmsg+' '+target.name+' :only a player, channel operator, or irc operator can end the game')
            end;
          end else begin
            send(mybot.idstr+' '+tok_privmsg+' '+target.name+' :no game in progress')
          end;
        end;
      end;
      firstcolchar..lastcolchar : begin
        g := tconnect4game(target.bottag[srvunit.num]);
        if assigned(target.bottag[srvunit.num]) and (length(m2)=2) then with g do begin

          if (length(s) = 2) and (currentplayer =sptr) then begin

            if checkpenalty then exit;

            row := -1; // it becomes 0 before first reaching the until
            repeat
              inc(row)
            until (row=lastrow) or (board[ord(m2[2])-ord('0'),row]=0);
            col := ord(m2[2])-ord('0');
            if board[col,row]=0 then begin

              if currentplayer=player1 then begin
                board[col,row]:=1;
              end else begin
                board[col,row]:=2;
              end;
            end else begin
              send(mybot.idstr+' '+tok_privmsg+' '+target.name+' :column full');
              exit;
            end;;
            won := false;

            // code to find a win
            currentplayernum := board[ord(m2[2])-ord('0'),row];
            for counter := 0 to 7 do begin
              changerow := changerowcoltable[counter,0];
              changecol := changerowcoltable[counter,1];
              wingrid[changecol,changerow] := 0;

              //debugout(inttostr(boardrc(col+(changecol*(wingrid[changecol,changerow]+1)),row+(changerow*(wingrid[changecol,changerow]+1)))));
              //delphi gives a might not have been initialized warning here as usual this is incorrect
              while boardrc(col+(changecol*(wingrid[changecol,changerow]+1)),row+(changerow*(wingrid[changecol,changerow]+1)))=currentplayernum do begin;
                inc(wingrid[changecol,changerow]);
              end;
            end;

            if wingrid[0,1]+wingrid[0,-1]+1 >= numberinrowtowin then won := true;
            if wingrid[1,1]+wingrid[-1,-1]+1 >= numberinrowtowin then won := true;
            if wingrid[1,0]+wingrid[-1,0]+1 >= numberinrowtowin then won := true;
            if wingrid[1,-1]+wingrid[-1,1]+1 >= numberinrowtowin then won := true;
            if not won then begin
              if currentplayer=player1 then begin
                currentplayer := player2;
              end else begin
                currentplayer := player1;
              end;
              //check for full board
              zerofound := false;
              for col := firstcol to lastcol do begin
                if board[col,lastrow]=0 then zerofound := true;

              end;
              if not zerofound then currentplayer := nil; // game drawn by full board


            end;
            drawboard(target,won);
            if won then begin
              endgame(target,currentplayer.name+' won!');
            end;
            if currentplayer=nil then begin
              endgame(target,'board is full');
            end;
          end;
        end;
      end;
    end;

  end;

end;

procedure tconnect4bot.userleavechannel(ch:tchannel;user:tuser);
begin
  if assigned(ch.bottag[srvunit.num]) then begin
    if (tconnect4game(ch.bottag[srvunit.num]).player1 =user) or (tconnect4game(ch.bottag[srvunit.num]).player2 =user) then begin
      endgame(ch,user.name+' left the channel')
    end;;
  end;
end;

procedure tconnect4bot.botleavechannel(ch:tchannel);
begin
  endgame(ch,mybot.name+' left the channel')
end;
procedure onstartup(tag:taddrint);
begin
  with tconnect4bot(tag) do begin
    if up then exit; //up will be set to true in invkickbot.onstartup


    optiontable_default(optiontable,maxoptiontable);

    optiontable_load(optiontable,maxoptiontable,configfile,cfgtag);
    //writeln(configfile);
    optiontable_save(optiontable,maxoptiontable,configfile,cfgtag);

    nick := locopt.nick;
    userid := locopt.userid;
    host := locopt.host;
    realname := locopt.info;
    invkickbot.onstartup(tag);

  end;
end;

procedure tconnect4bot.setupsrvunit;
begin
  inherited setupsrvunit;
  srvunit.onchanmsg := onchanmsg;
  srvunit.onprivmsg := onprivmsg;
  srvunit.onstartup := onstartup;
  srvunit.name := 'CONNECT4';
end;

constructor tconnect4bot.create;
begin
  inherited create;


end;

initialization
  tconnect4bot.create;
end.

